/*
  nokia_gps_gio
  description: Nokia/GPS sample clock with time zone knob
  Version: 2
  By: Tim Mikkelsen
  Created: 4/06/2016
  Updated: 5/04/2016

  This Arduino code/project is for an LCD clock using a
  NEMA style GPS and Nokia cell phone LCD display

  Hardware:
  - Arduino - Arduino Pro Mini 8Mhz/3.3V      (Sparkfun DEV-11114)
  - Nokia 5110 graphic 84x48 LCD display      (Sparkfun LCD-10168)
  - GPS Receiver - GP-20U7                    (Sparkfun GPS-13470)
  - one 10k trimpot                           (Sparkfun COM-09806)
  - FTDI Basic Breakout (to program Arduino)  (Sparkfun DEV-09873)
  - 6v wall power supply

  Additional libraries and code segments:
  - timezone adjustment code (by Jinseok Jeon)
  - Nokia-5110 graphic LCD code (from the Arduino community)
  - TinyGPS++ GPS library (by Mikal Hart)
  - SoftwareSerial UART library (by Mikal Hart)

  Note the use of smartDelay to ensure not losing data - from the
  TinyGPS++ example.  There are still occasional checksum errors
  and overflows, but the data does get read
*/

// libraries  -----------------------------------------------------------------
#include <TinyGPS++.h>
#include <SoftwareSerial.h>
#include <string.h>
#include <ctype.h>

// constants  -----------------------------------------------------------------
static const int TZPotPin   = 0;          // timezone pot pin defintion     A0
static const int WorkingPin = 13;         // gps working pin definition     D13
static const int RXPin      = 8;          // gps serial Rx pin definition   D8
static const int TXPin      = 9;          // gps serial Tx pin definition   D9
static const uint32_t GPSBaud = 9600;     // gps baud rate
// static const uint32_t DbgBaud = 9600;  // debug serial monitor baud rate

// LCD pin definitions --------------------------------------------------------
static const int PIN_SCE   = 7;
static const int PIN_RESET = 6;
static const int PIN_DC    = 5;
static const int PIN_SDIN  = 4;
static const int PIN_SCLK  = 3;
static const int LCD_C     = LOW;
static const int LCD_D     = HIGH;
static const int LCD_X     = 84;
static const int LCD_Y     = 48;

// LCD character bitmaps ------------------------------------------------------
static const byte ASCII[][5] =
{
  {0x00, 0x00, 0x00, 0x00, 0x00} // 20
  , {0x00, 0x00, 0x5f, 0x00, 0x00} // 21 !
  , {0x00, 0x07, 0x00, 0x07, 0x00} // 22 "
  , {0x14, 0x7f, 0x14, 0x7f, 0x14} // 23 #
  , {0x24, 0x2a, 0x7f, 0x2a, 0x12} // 24 $
  , {0x23, 0x13, 0x08, 0x64, 0x62} // 25 %
  , {0x36, 0x49, 0x55, 0x22, 0x50} // 26 &
  , {0x00, 0x05, 0x03, 0x00, 0x00} // 27 '
  , {0x00, 0x1c, 0x22, 0x41, 0x00} // 28 (
  , {0x00, 0x41, 0x22, 0x1c, 0x00} // 29 )
  , {0x14, 0x08, 0x3e, 0x08, 0x14} // 2a *
  , {0x08, 0x08, 0x3e, 0x08, 0x08} // 2b +
  , {0x00, 0x50, 0x30, 0x00, 0x00} // 2c ,
  , {0x08, 0x08, 0x08, 0x08, 0x08} // 2d -
  , {0x00, 0x60, 0x60, 0x00, 0x00} // 2e .
  , {0x20, 0x10, 0x08, 0x04, 0x02} // 2f /
  , {0x3e, 0x51, 0x49, 0x45, 0x3e} // 30 0
  , {0x00, 0x42, 0x7f, 0x40, 0x00} // 31 1
  , {0x42, 0x61, 0x51, 0x49, 0x46} // 32 2
  , {0x21, 0x41, 0x45, 0x4b, 0x31} // 33 3
  , {0x18, 0x14, 0x12, 0x7f, 0x10} // 34 4
  , {0x27, 0x45, 0x45, 0x45, 0x39} // 35 5
  , {0x3c, 0x4a, 0x49, 0x49, 0x30} // 36 6
  , {0x01, 0x71, 0x09, 0x05, 0x03} // 37 7
  , {0x36, 0x49, 0x49, 0x49, 0x36} // 38 8
  , {0x06, 0x49, 0x49, 0x29, 0x1e} // 39 9
  , {0x00, 0x36, 0x36, 0x00, 0x00} // 3a :
  , {0x00, 0x56, 0x36, 0x00, 0x00} // 3b ;
  , {0x08, 0x14, 0x22, 0x41, 0x00} // 3c <
  , {0x14, 0x14, 0x14, 0x14, 0x14} // 3d =
  , {0x00, 0x41, 0x22, 0x14, 0x08} // 3e >
  , {0x02, 0x01, 0x51, 0x09, 0x06} // 3f ?
  , {0x32, 0x49, 0x79, 0x41, 0x3e} // 40 @
  , {0x7e, 0x11, 0x11, 0x11, 0x7e} // 41 A
  , {0x7f, 0x49, 0x49, 0x49, 0x36} // 42 B
  , {0x3e, 0x41, 0x41, 0x41, 0x22} // 43 C
  , {0x7f, 0x41, 0x41, 0x22, 0x1c} // 44 D
  , {0x7f, 0x49, 0x49, 0x49, 0x41} // 45 E
  , {0x7f, 0x09, 0x09, 0x09, 0x01} // 46 F
  , {0x3e, 0x41, 0x49, 0x49, 0x7a} // 47 G
  , {0x7f, 0x08, 0x08, 0x08, 0x7f} // 48 H
  , {0x00, 0x41, 0x7f, 0x41, 0x00} // 49 I
  , {0x20, 0x40, 0x41, 0x3f, 0x01} // 4a J
  , {0x7f, 0x08, 0x14, 0x22, 0x41} // 4b K
  , {0x7f, 0x40, 0x40, 0x40, 0x40} // 4c L
  , {0x7f, 0x02, 0x0c, 0x02, 0x7f} // 4d M
  , {0x7f, 0x04, 0x08, 0x10, 0x7f} // 4e N
  , {0x3e, 0x41, 0x41, 0x41, 0x3e} // 4f O
  , {0x7f, 0x09, 0x09, 0x09, 0x06} // 50 P
  , {0x3e, 0x41, 0x51, 0x21, 0x5e} // 51 Q
  , {0x7f, 0x09, 0x19, 0x29, 0x46} // 52 R
  , {0x46, 0x49, 0x49, 0x49, 0x31} // 53 S
  , {0x01, 0x01, 0x7f, 0x01, 0x01} // 54 T
  , {0x3f, 0x40, 0x40, 0x40, 0x3f} // 55 U
  , {0x1f, 0x20, 0x40, 0x20, 0x1f} // 56 V
  , {0x3f, 0x40, 0x38, 0x40, 0x3f} // 57 W
  , {0x63, 0x14, 0x08, 0x14, 0x63} // 58 X
  , {0x07, 0x08, 0x70, 0x08, 0x07} // 59 Y
  , {0x61, 0x51, 0x49, 0x45, 0x43} // 5a Z
  , {0x00, 0x7f, 0x41, 0x41, 0x00} // 5b [
  , {0x02, 0x04, 0x08, 0x10, 0x20} // 5c ¥
  , {0x00, 0x41, 0x41, 0x7f, 0x00} // 5d ]
  , {0x04, 0x02, 0x01, 0x02, 0x04} // 5e ^
  , {0x40, 0x40, 0x40, 0x40, 0x40} // 5f _
  , {0x00, 0x01, 0x02, 0x04, 0x00} // 60 `
  , {0x20, 0x54, 0x54, 0x54, 0x78} // 61 a
  , {0x7f, 0x48, 0x44, 0x44, 0x38} // 62 b
  , {0x38, 0x44, 0x44, 0x44, 0x20} // 63 c
  , {0x38, 0x44, 0x44, 0x48, 0x7f} // 64 d
  , {0x38, 0x54, 0x54, 0x54, 0x18} // 65 e
  , {0x08, 0x7e, 0x09, 0x01, 0x02} // 66 f
  , {0x0c, 0x52, 0x52, 0x52, 0x3e} // 67 g
  , {0x7f, 0x08, 0x04, 0x04, 0x78} // 68 h
  , {0x00, 0x44, 0x7d, 0x40, 0x00} // 69 i
  , {0x20, 0x40, 0x44, 0x3d, 0x00} // 6a j
  , {0x7f, 0x10, 0x28, 0x44, 0x00} // 6b k
  , {0x00, 0x41, 0x7f, 0x40, 0x00} // 6c l
  , {0x7c, 0x04, 0x18, 0x04, 0x78} // 6d m
  , {0x7c, 0x08, 0x04, 0x04, 0x78} // 6e n
  , {0x38, 0x44, 0x44, 0x44, 0x38} // 6f o
  , {0x7c, 0x14, 0x14, 0x14, 0x08} // 70 p
  , {0x08, 0x14, 0x14, 0x18, 0x7c} // 71 q
  , {0x7c, 0x08, 0x04, 0x04, 0x08} // 72 r
  , {0x48, 0x54, 0x54, 0x54, 0x20} // 73 s
  , {0x04, 0x3f, 0x44, 0x40, 0x20} // 74 t
  , {0x3c, 0x40, 0x40, 0x20, 0x7c} // 75 u
  , {0x1c, 0x20, 0x40, 0x20, 0x1c} // 76 v
  , {0x3c, 0x40, 0x30, 0x40, 0x3c} // 77 w
  , {0x44, 0x28, 0x10, 0x28, 0x44} // 78 x
  , {0x0c, 0x50, 0x50, 0x50, 0x3c} // 79 y
  , {0x44, 0x64, 0x54, 0x4c, 0x44} // 7a z
  , {0x00, 0x08, 0x36, 0x41, 0x00} // 7b {
  , {0x00, 0x00, 0x7f, 0x00, 0x00} // 7c |
  , {0x00, 0x41, 0x36, 0x08, 0x00} // 7d }
  , {0x10, 0x08, 0x08, 0x10, 0x08} // 7e ←
  , {0x78, 0x46, 0x41, 0x46, 0x78} // 7f →
};

// variables  -----------------------------------------------------------------
int gpsYear;
int gpsMonth;
int gpsDay;
int gpsHour;
int TZPotValue;                           // Timezone pot input on A0 input
int TimeZone;                             // Timezone
byte gpsMin;
byte gpsSec;
String TimeStr;
String DateStr;
String NameOrMsg = "Tim's clock";         // Name or message - max of 12 chars
String LatStr;                            // latitude
String LngStr;                            // longitude
String AltStr;                            // altitude

// Daylight Savings tables through 2032 ---------------------------------------
int DSTbegin[] = {                        // DST 2015 - 2032 in Canada and US
  308, 313, 312, 311, 310, 308, 314, 313, 312,            // (MMdd coding)
  310, 309, 308, 314, 312, 311, 310, 309, 314
};
int DSTend[] = {                          // DST 2015 - 2032 in Canada and US
  1101, 1106, 1105, 1104, 1103, 1101, 1107, 1106, 1105,   // (MMdd coding)
  1103, 1102, 1101, 1107, 1105, 1104, 1103, 1102, 1107
};
int DaysAMonth[] = {                      //number of days a month
  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};


// create devices  ------------------------------------------------------------
TinyGPSPlus gps;                          // TinyGPS++ object
SoftwareSerial ss(RXPin, TXPin);          // serial connection to the GPS device

// This custom version of delay() ensures that the gps object  ----------------
// is being "fed".
static void smartDelay(unsigned long ms)
{
  unsigned long start = millis();
  do
  {
    while (ss.available())
      gps.encode(ss.read());
  } while (millis() - start < ms);
}

// Nokia 5110 graphical LCD subroutines  --------------------------------------
void LcdCharacter(char character)
{
  LcdWrite(LCD_D, 0x00);
  for (int index = 0; index < 5; index++)
  {
    LcdWrite(LCD_D, ASCII[character - 0x20][index]);
  }
  LcdWrite(LCD_D, 0x00);
}

void LcdClear(void)
{
  for (int index = 0; index < LCD_X * LCD_Y / 8; index++)
  {
    LcdWrite(LCD_D, 0x00);
  }
}

void LcdInitialize(void)
{
  pinMode(PIN_SCE, OUTPUT);
  pinMode(PIN_RESET, OUTPUT);
  pinMode(PIN_DC, OUTPUT);
  pinMode(PIN_SDIN, OUTPUT);
  pinMode(PIN_SCLK, OUTPUT);
  digitalWrite(PIN_RESET, LOW);

  digitalWrite(PIN_RESET, HIGH);
  LcdWrite(LCD_C, 0x21 );  // LCD Extended Commands.
  LcdWrite(LCD_C, 0xBf );  // Set LCD Vop (Contrast).
  LcdWrite(LCD_C, 0x04 );  // Set Temp coefficent. //0x04
  LcdWrite(LCD_C, 0x14 );  // LCD bias mode 1:48.  //0x13
  LcdWrite(LCD_C, 0x0C );  // LCD in normal mode.
  LcdWrite(LCD_C, 0x20 );
  LcdWrite(LCD_C, 0x0C );
}

void LcdSetCursor(int x, int y)
// x - range: 0 to 83
// y - range: 0 to 5
{
  LcdWrite( 0, 0x80 | x);  // Column.
  LcdWrite( 0, 0x40 | y);  // Row.
  smartDelay(0);
}

void LcdString(String aString)
{
  for (int i = 0; i < aString.length() ; i++) LcdCharacter(aString[i]);
}

void LcdWrite(byte dc, byte data)
{
  digitalWrite(PIN_DC, dc);
  digitalWrite(PIN_SCE, LOW);
  shiftOut(PIN_SDIN, PIN_SCLK, MSBFIRST, data);
  digitalWrite(PIN_SCE, HIGH);
}

// set up SmartSerial
void setup()
{
  // set up GPS serial interface  ---------------------------------------------
  ss.begin(GPSBaud);

  // set up LCD  --------------------------------------------------------------
  LcdInitialize();
  LcdClear();
  // LcdString("starting up");
}


// main loop of program
void loop()
{
  // display information every time a new stream is correctly read  -----------
  while (ss.available() > 0)
    if (gps.encode(ss.read()))
      displayInfo();
  smartDelay(0);
}

void displayInfo()
{

  // make sure gps data is good  ----------------------------------------------
  if (gps.date.day() == 0 &&
      gps.date.month() == 0)
  {
    LcdClear();
    LcdString("acquiring gps");
    digitalWrite(WorkingPin, LOW);

  }
  else {

    // clear out the time and date strings
    TimeStr = "";
    DateStr = "";

    // turn on LED on pint 13 to show functioning
    digitalWrite(WorkingPin, HIGH);

    // get actual UTC gps data
    gpsYear = gps.date.year();
    gpsMonth = gps.date.month();
    gpsDay = gps.date.day();
    gpsHour = gps.time.hour();

    // get timezone input -----------------------------------------------------
    TZPotValue = analogRead(TZPotPin);
    TimeZone = ((TZPotValue - 512) / 40);
    if (TimeZone < -12) TimeZone = -12;
    if (TimeZone > 12) TimeZone = 12;

    // leap year check  -------------------------------------------------------
    if (gps.date.year() % 4 == 0) DaysAMonth[1] = 29;
    else DaysAMonth[1] = 28;

    // time zone adjustment  --------------------------------------------------
    gpsHour += TimeZone;

    // daylight savings time adjustments  -------------------------------------
    if (gpsMonth * 100 + gpsDay >= DSTbegin[gpsYear - 13] &&
        gpsMonth * 100 + gpsDay < DSTend[gpsYear - 13]) gpsHour += 1;
    if (gpsHour < 0)
    {
      gpsHour += 24;
      gpsDay -= 1;
      if (gpsDay < 1)
      {
        if (gpsMonth == 1)
        {
          gpsMonth = 12;
          gpsYear -= 1;
        }
        else
        {
          gpsMonth -= 1;
        }
        gpsDay = DaysAMonth[gpsMonth - 1];
      }
    }
    if (gpsHour >= 24)
    {
      gpsHour -= 24;
      gpsDay += 1;
      if (gpsDay > DaysAMonth[gpsMonth - 1])
      {
        gpsDay = 1;
        gpsMonth += 1;
        if (gpsMonth > 12) gpsYear += 1;
      }
    }

    // print name or message --------------------------------------------------
    smartDelay(0);
    LcdSetCursor(0, 0);
    // date checks  -----------------------------------------------------------
    // note there are only 12 characters for the message     123456789012
    if      ((gpsMonth == 1)  && (gpsDay == 1))   LcdString("New Year's  ");
    else if ((gpsMonth == 4)  && (gpsDay == 22))  LcdString("Earth Day   ");
    else if ((gpsMonth == 6)  && (gpsDay == 14))  LcdString("Flag Day    ");
    else if ((gpsMonth == 7)  && (gpsDay == 4))   LcdString("4th of July ");
    else if ((gpsMonth == 12) && (gpsDay == 25))  LcdString("Christmas   ");
    else                                          LcdString(NameOrMsg);


    // build the date string --------------------------------------------------
    smartDelay(0);

    if (gpsMonth < 10) DateStr = "0";
    DateStr += gpsMonth;
    DateStr += "/";
    if (gpsDay < 10) DateStr += "0";
    DateStr += gpsDay;
    DateStr += "/";
    gpsYear = gpsYear % 100;
    if (gpsYear < 10) DateStr += "0";
    DateStr += gpsYear;

    LcdSetCursor(0, 9);
    LcdString("    " + DateStr);


    // build the time string --------------------------------------------------
    smartDelay(0);

    if (gpsHour < 10) TimeStr = "0";
    TimeStr += gpsHour;
    TimeStr += ":";

    if (gps.time.minute() < 10) TimeStr += "0";
    TimeStr += gps.time.minute();
    TimeStr += ":";

    if (gps.time.second() < 10) TimeStr += "0";
    TimeStr += gps.time.second();

    LcdSetCursor(0, 18);
    LcdString("    " + TimeStr);                          // display the time


    // build the latitude and longitude string --------------------------------
    smartDelay(0);

    LatStr = String(gps.location.lat());
    LngStr = String(gps.location.lng());
    while ( LatStr.length() != LngStr.length() ) {
      if ( LatStr.length() < LngStr.length() ) LatStr = " " + LatStr;
      else if ( LatStr.length() > LngStr.length() ) LngStr = " " + LngStr;
    }

    LcdSetCursor(0, 27);
    LcdString("LAT: " + LatStr);                // display the latitude

    LcdSetCursor(0, 36);
    LcdString("LNG: " + LngStr);                // display the longitude

    // print altitude string  -------------------------------------------------
    LcdSetCursor(0, 45);
    AltStr = String(int(gps.altitude.feet()));  // get altitude as an int
    while ( AltStr.length() < 4 ) {
      AltStr = " " + AltStr;
    }

    LcdString("ALT: " + AltStr + " Ft");                // display the altitude
    smartDelay(0);



  }
}



